package io.reisub.openosrs.tempoross;

import com.google.common.collect.ImmutableSet;
import com.google.inject.Provides;
import io.reisub.openosrs.tempoross.tasks.*;
import io.reisub.openosrs.util.Task;
import io.reisub.openosrs.util.Util;
import io.reisub.openosrs.util.tasks.Run;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import net.runelite.api.*;
import net.runelite.api.events.*;
import net.runelite.client.config.ConfigManager;
import net.runelite.client.eventbus.Subscribe;
import net.runelite.client.plugins.PluginDependency;
import net.runelite.client.plugins.PluginDescriptor;
import net.runelite.client.plugins.iutils.game.iObject;
import net.runelite.client.plugins.iutils.game.iWidget;
import net.runelite.client.plugins.iutils.iUtils;
import net.runelite.client.plugins.iutils.scripts.iScript;
import org.pf4j.Extension;

import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static net.runelite.api.AnimationID.IDLE;

@Extension
@PluginDependency(Util.class)
@PluginDependency(iUtils.class)
@PluginDescriptor(
	name = "Chaos Tempoross",
	description = "Plays the Tempoross minigame",
	enabledByDefault = false
)
@Slf4j
public class Tempoross extends iScript {
	private static final String WAVE_INCOMING_MESSAGE = "a colossal wave closes in...";
	private static final String WAVE_END_SAFE = "as the wave washes over you";
	private static final String WAVE_END_DANGEROUS = "the wave slams into you";
	private static final String TEMPOROSS_VULNERABLE_MESSAGE = "tempoross is vulnerable";

	private static final int VARB_IS_TETHERED = 11895;

	private static final int TEMPOROSS_REGION = 12078;
	private static final int UNKAH_REWARD_POOL_REGION = 12588;
	private static final int UNKAH_BOAT_REGION = 12332;

	private static final int FIRE_SPREAD_MILLIS = 24000;
	private static final int FIRE_SPAWN_MILLIS = 9600;
	private static final int FIRE_SPREADING_SPAWN_MILLIS = 1200;
	private static final int WAVE_IMPACT_MILLIS = 7800;

	public static final int FIRE_ID = 37582;
	public static final int DOUSING_FIRE_ANIM_ID = 2771;

	private final Set<Integer> TEMPOROSS_GAMEOBJECTS = ImmutableSet.of(
			FIRE_ID, NullObjectID.NULL_41006, NullObjectID.NULL_41007, NullObjectID.NULL_41352,
			NullObjectID.NULL_41353, NullObjectID.NULL_41354, NullObjectID.NULL_41355, ObjectID.DAMAGED_MAST_40996,
			ObjectID.DAMAGED_MAST_40997, ObjectID.DAMAGED_TOTEM_POLE, ObjectID.DAMAGED_TOTEM_POLE_41011);

	//41006 = shadow before fire is burning
	//41007 = shadow just before fire is jumping over to a next spot
	//41354/41355 = a totem to grapple on to
	//41352/41353 = a mast to grapple on to
	//41010/41011 = a totem that is broken
	//40996/40997 = a broken mast

	@Getter
	private final Map<GameObject, TemporossObject> temporossObjects = new HashMap<>();

	@Getter
	private boolean waveIncoming;

	@Getter
	private int phase = 1;

	@Getter
	private int playersReady;

	@Getter
	private int energy;

	@Getter
	private int essence;

	@Getter
	private int stormIntensity;

	@Getter
	private Activity currentActivity;

	@Getter
	private Instant lastActionTime;

	private List<Task> tasks;

	@SuppressWarnings("unused")
	@Provides
    Config provideConfig(ConfigManager configManager) {
		return configManager.getConfig(Config.class);
	}

	@Override
	protected void loop() {
		for (Task t : tasks) {
			if (t.validate()) {
				log.info(t.getStatus());
				t.execute();
				break;
			}
		}

		game.sleepDelay();
	}

	@Override
	protected void onStart() {
		log.info("Starting Chaos Tempoross");

		Run runTask = injector.getInstance(Run.class);
		runTask.setInterval(70, 95);

		tasks = new ArrayList<>();
		tasks.add(runTask);
		tasks.add(injector.getInstance(EnterBoat.class));
		tasks.add(injector.getInstance(FillBuckets.class));
		tasks.add(injector.getInstance(LeaveBoat.class));
		tasks.add(injector.getInstance(Repair.class));
		tasks.add(injector.getInstance(Tether.class));
		tasks.add(injector.getInstance(Attack.class));
		tasks.add(injector.getInstance(Stock.class));
		tasks.add(injector.getInstance(Fish.class));
		tasks.add(injector.getInstance(Cook.class));
	}

	@Override
	protected void onStop() {
		log.info("Stopping Chaos Tempoross");
		if (tasks != null) {
			tasks.clear();
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onConfigButtonPressed(ConfigButtonClicked configButtonClicked) {
		if (configButtonClicked.getKey().equals("startButton")) {
			execute();
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onGameTick(GameTick event) {
		if (isInDesert() || isOnBoat()) {
			playersReady = parseWidget(687, 3);
		} else if (isInTemperossArea()) {
			energy = parseWidget(437, 35);
			essence = parseWidget(437, 45);
			stormIntensity = parseWidget(437, 55);
		}

		checkActionTimeout();
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onGameObjectSpawned(GameObjectSpawned event) {
		GameObject gameObject = event.getGameObject();

		if (!TEMPOROSS_GAMEOBJECTS.contains(gameObject.getId())) return;

		TemporossObject temporossObject;

		switch (event.getGameObject().getId()) {
			case FIRE_ID:
				temporossObject = new TemporossObject(gameObject, Instant.now(), FIRE_SPREAD_MILLIS);
				break;
			case NullObjectID.NULL_41006:
				temporossObject = new TemporossObject(gameObject, Instant.now(), FIRE_SPAWN_MILLIS);
				break;
			case NullObjectID.NULL_41007:
				temporossObject = new TemporossObject(gameObject, Instant.now(), FIRE_SPREADING_SPAWN_MILLIS);
				break;
			default:
				temporossObject = new TemporossObject(gameObject, Instant.now(), WAVE_IMPACT_MILLIS);
		}

		temporossObjects.put(gameObject, temporossObject);
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onGameObjectDespawned(GameObjectDespawned gameObjectDespawned) {
		int id = gameObjectDespawned.getGameObject().getId();

		if (id == FIRE_ID) {
			if (currentActivity == Activity.DOUSING_FIRE) {
				setActivity(Activity.IDLE);
			}
		}

		Set<Integer> brokenMastsTotems = ImmutableSet.of(
				ObjectID.DAMAGED_MAST_40996,
				ObjectID.DAMAGED_MAST_40997,
				ObjectID.DAMAGED_TOTEM_POLE,
				ObjectID.DAMAGED_TOTEM_POLE_41011
		);

		if (brokenMastsTotems.contains(id)) {
			if (currentActivity == Activity.REPAIRING_MAST) {
				setActivity(Activity.IDLE);
			}
		}

		temporossObjects.remove(gameObjectDespawned.getGameObject());
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onChatMessage(ChatMessage chatMessage) {
		if (chatMessage.getType() != ChatMessageType.GAMEMESSAGE) return;

		String message = chatMessage.getMessage().toLowerCase();
		if (message.contains(WAVE_INCOMING_MESSAGE)) {
			waveIncoming = true;
		} else if (message.contains(WAVE_END_SAFE) || message.contains(WAVE_END_DANGEROUS)) {
			if (message.contains(WAVE_END_SAFE)) {
				setActivity(Activity.IDLE);
			}
			waveIncoming = false;
		} else if (message.contains(TEMPOROSS_VULNERABLE_MESSAGE)) {
			phase++;
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onAnimationChanged(AnimationChanged event) {
		if (game.client().getGameState() != GameState.LOGGED_IN) return;

		if (event.getActor() != game.client().getLocalPlayer()) return;

		int animId = game.localPlayer().animation();
		switch (animId) {
			case AnimationID.FISHING_BAREHAND:
			case AnimationID.FISHING_BAREHAND_WINDUP_1:
			case AnimationID.FISHING_BAREHAND_WINDUP_2:
				setActivity(Activity.FISHING);
			case AnimationID.COOKING_RANGE:
				iObject shrine = game.objects().withName("Shrine").nearest();
				if (shrine != null && shrine.position().distanceTo(game.localPlayer().position()) <= 3) {
					setActivity(Activity.COOKING);
				} else {
					setActivity(Activity.STOCKING_CANNON);
				}
			case DOUSING_FIRE_ANIM_ID:
				setActivity(Activity.DOUSING_FIRE);
			case AnimationID.CONSTRUCTION:
			case AnimationID.CONSTRUCTION_IMCANDO:
				setActivity(Activity.REPAIRING_MAST);
			case AnimationID.USING_GILDED_ALTAR:
				setActivity(Activity.TETHERING_MAST);
			case AnimationID.LOOKING_INTO:
				setActivity(Activity.FILLING_BUCKETS);
			case 0:
				setActivity(Activity.ATTACKING);
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onItemContainerChanged(ItemContainerChanged event) {
		final ItemContainer container = event.getItemContainer();

		int rawFish = (int) game.inventory().withId(ItemID.RAW_HARPOONFISH).count();
		int cookedFish = (int) game.inventory().withId(ItemID.HARPOONFISH).count();
		int emptyBuckets = (int) game.inventory().withId(ItemID.BUCKET).count();

		if (rawFish == 0 && currentActivity == Activity.COOKING) {
			setActivity(Activity.IDLE);
		} else if (game.inventory().full() && currentActivity == Activity.FISHING) {
			setActivity(Activity.IDLE);
		} else if (cookedFish == 0 && currentActivity == Activity.STOCKING_CANNON) {
			setActivity(Activity.IDLE);
		} else if (emptyBuckets == 0 && currentActivity == Activity.FILLING_BUCKETS) {
			setActivity(Activity.IDLE);
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onNpcSpawned(NpcSpawned npcSpawned) {
		if (npcSpawned.getNpc().getId() == NpcID.FISHING_SPOT_10569
				&& !game.inventory().full()
				&& (currentActivity == Activity.COOKING || currentActivity == Activity.FISHING)) {
			setActivity(Activity.IDLE);
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onVarbitChanged(VarbitChanged event) {
		if (game.client().getVarbitValue(VARB_IS_TETHERED) > 0) {
			setActivity(Activity.TETHERING_MAST);
		}

		if (game.client().getVarbitValue(VARB_IS_TETHERED) <= 0 && currentActivity == Activity.TETHERING_MAST) {
			setActivity(Activity.IDLE);
		}
	}

	public boolean isOnBoat() {
		return game.localPlayer().position().regionID() == UNKAH_BOAT_REGION;
	}

	public boolean isInDesert() {
		return game.localPlayer().position().regionID() == UNKAH_REWARD_POOL_REGION;
	}

	public boolean isInTemperossArea() {
		return game.localPlayer().position().regionID() == TEMPOROSS_REGION;
	}

	private int parseWidget(int group, int file) {
		iWidget widget = game.widget(group, file);
		if (widget == null || widget.text().equals("")) return 0;

		Pattern regex = Pattern.compile("\\d+|None");
		Matcher matcher = regex.matcher(widget.text());

		if (matcher.find()) {
			String match = matcher.group(0);
			if (match.equals("None")) return 0;

			return Integer.parseInt(match);
		}

		return 0;
	}

	private void setActivity(Activity action) {
		currentActivity = action;

		if (action != Activity.IDLE) {
			lastActionTime = Instant.now();
		}
	}

	private void checkActionTimeout() {
		if (currentActivity == Activity.IDLE) return;

		int animId = game.localPlayer().animation();
		if (animId != IDLE || lastActionTime == null) return;

		Duration timeout = Duration.ofSeconds(3);
		Duration sinceAction = Duration.between(lastActionTime, Instant.now());

		if (sinceAction.compareTo(timeout) >= 0) {
			setActivity(Activity.IDLE);
		}
	}
}