package io.reisub.openosrs.util;

import io.reisub.openosrs.util.enums.Activity;
import io.reisub.openosrs.util.tasks.KittenTask;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import net.runelite.api.ChatMessageType;
import net.runelite.api.GameState;
import net.runelite.api.InventoryID;
import net.runelite.api.Skill;
import net.runelite.api.events.*;
import net.runelite.client.eventbus.Subscribe;
import net.runelite.client.plugins.iutils.game.iPlayer;
import net.runelite.client.plugins.iutils.scripts.iScript;

import javax.inject.Inject;
import java.time.Duration;
import java.time.Instant;
import java.util.*;

import static net.runelite.api.AnimationID.IDLE;

@Slf4j
public abstract class CScript extends iScript {
    @Inject
    protected Calculations calc;

    protected final List<Task> tasks = new ArrayList<>();
    protected KittenTask kittenTask;

    @Getter
    protected Activity currentActivity;

    @Getter
    protected Activity previousActivity;

    @Getter
    @Setter
    private Instant lastHop;

    protected Instant lastLogin = Instant.EPOCH;
    protected Instant lastActionTime = Instant.EPOCH;
    protected Duration lastActionTimeout = Duration.ofSeconds(3);
    protected Instant lastExperience = Instant.EPOCH;
    protected Instant lastInventoryChange = Instant.EPOCH;
    protected final Map<Skill, Activity> idleCheckSkills = new HashMap<>();
    protected boolean idleCheckInventoryChange = false;

    @Override
    protected void loop() {
        for (Task t : tasks) {
            if (t.validate()) {
                log.info(t.getStatus());
                t.execute();
                break;
            }
        }

        checkActionTimeout();
        game.sleepDelay();
    }

    @Override
    protected void onStart() {
        log.info("Starting " + this.getName());

        previousActivity = Activity.IDLE;
        currentActivity = Activity.IDLE;

        kittenTask = KittenTask.getInstance(injector);
    }

    @Override
    protected void onStop() {
        log.info("Stopping " + this.getName());
        tasks.clear();

        previousActivity = Activity.IDLE;
        currentActivity = Activity.IDLE;

        KittenTask.handleKitten = false;
    }

    protected final <T extends Task> void addTask(Class<T> type) {
        tasks.add(injector.getInstance(type));
    }

    @SuppressWarnings("unused")
    @Subscribe
    private void onConfigButtonPressed(ConfigButtonClicked event) {
        String name = this.getName().replaceAll(" ", "").toLowerCase(Locale.ROOT);

        if (event.getGroup().equals(name) && event.getKey().equals("startButton")) {
            execute();
        }
    }

    @SuppressWarnings("unused")
    @Subscribe
    private void onChatMessage(ChatMessage event) {
        if (kittenTask != null) {
            kittenTask.onChatMessage(event);
        }

        if (event.getType() == ChatMessageType.GAMEMESSAGE) {
            if (event.getMessage().startsWith("Congratulations, you've just advanced your")) {
                setActivity(Activity.IDLE);
            }
        }
    }

    @SuppressWarnings("unused")
    @Subscribe
    private void onStatChanged(StatChanged event) {
        for (Skill skill : idleCheckSkills.keySet()) {
            if (event.getSkill() == skill) {
                setActivity(idleCheckSkills.get(skill));
                lastExperience = Instant.now();
            }
        }
    }
    @SuppressWarnings("unused")
    @Subscribe
    private void onItemContainerChanged(ItemContainerChanged event) {
        if (!isLoggedIn() || event.getItemContainer() != game.client().getItemContainer(InventoryID.INVENTORY)) return;

        if (idleCheckInventoryChange) {
            lastInventoryChange = Instant.now();
        }
    }

    @Subscribe
    @SuppressWarnings("unused")
    private void onGameStateChanged(GameStateChanged event) {
        if (event.getGameState() == GameState.LOGGED_IN) {
            lastLogin = Instant.now();
        }
    }

    public void setActivity(Activity action) {
        if (action == Activity.IDLE && currentActivity != Activity.IDLE) {
            previousActivity = currentActivity;
        }

        currentActivity = action;

        if (action != Activity.IDLE) {
            lastActionTime = Instant.now();
        }
    }

    protected void checkActionTimeout() {
        if (currentActivity == Activity.IDLE) return;

        if (Duration.between(lastExperience, Instant.now()).compareTo(lastActionTimeout) < 0) return;

        if (Duration.between(lastInventoryChange, Instant.now()).compareTo(lastActionTimeout) < 0) return;

        int animId = game.localPlayer().animation();
        if (animId != IDLE || lastActionTime == null) return;

        Duration sinceAction = Duration.between(lastActionTime, Instant.now());

        if (sinceAction.compareTo(lastActionTimeout) >= 0) {
            setActivity(Activity.IDLE);
        }
    }

    public final boolean isLoggedIn() {
        return game.client() != null && game.client().getGameState() == GameState.LOGGED_IN;
    }

    public final boolean isLoggedInForLongerThan(Duration duration) {
        return Duration.between(lastLogin, Instant.now()).compareTo(duration) >= 0;
    }

    public final boolean isInRegion(int regionId) {
        iPlayer player = game.localPlayer();
        return player != null
                && player.position().regionID() == regionId;
    }

    public final void stop(String message) {
        game.utils.sendGameMessage(message);
        this.execute();
    }
}
