package io.reisub.openosrs.util;

import javax.inject.Singleton;
import java.util.Random;

@Singleton
public class Calculations {
    private final Random random = new Random();

    /**
     * Get a random number in the interval [min, max[.
     *
     * @param min inclusive minimum value
     * @param max exclusive maximum value
     * @return random number min <= n < max
     */
    public int random(int min, int max) {
        if (max <= min) return min;
        return random.nextInt(max - min) + min;
    }

    /**
     * Get a random number in the interval [min, max[.
     *
     * @param min inclusive minimum value
     * @param max exclusive maximum value
     * @return random number min <= n < max
     */
    public double random(double min, double max) {
        if (max <= min) return min;
        return min + random.nextDouble() * (max - min);
    }

    /**
     * Get a random Gaussian distributed number around a mean using deviation.
     *
     * @param mean mean value
     * @param deviation deviation from the mean
     * @param weighted if true the distribution will be weighted to the mean and right side (curve shifted to the right)
     * @return random number min <= n <= max according to normal distribution
     */
    public int randomGauss(int mean, int deviation, boolean weighted) {
        double v;

        if (weighted) {
            v = -Math.log(Math.abs(random.nextGaussian())) * deviation + mean;
        } else {
            v = random.nextGaussian() * deviation + mean;
        }

        return (int) v;
    }

    /**
     * Get a random Gaussian distributed number around a mean using deviation.
     *
     * @param mean mean value
     * @param deviation deviation from the mean
     * @return random number min <= n <= max according to normal distribution
     */
    public int randomGauss(int mean, int deviation) {
        return randomGauss(mean, deviation, false);
    }

    /**
     * Get a random Gaussian distributed number within an interval [min, max] around a mean using deviation.
     *
     * @param min inclusive minimum value
     * @param max inclusive maximum value
     * @param mean mean value
     * @param deviation deviation from the mean
     * @param weighted if true the distribution will be weighted to the mean and right side (curve shifted to the right)
     * @return random number min <= n <= max according to normal distribution
     */
    public int randomGauss(int min, int max, int mean, int deviation, boolean weighted) {
        return clamp(randomGauss(mean, deviation, weighted), min, max);
    }

    /**
     * Get a random Gaussian distributed number within an interval [min, max] around a mean using deviation.
     *
     * @param min inclusive minimum value
     * @param max inclusive maximum value
     * @param mean mean value
     * @param deviation deviation from the mean
     * @return random number min <= n <= max according to normal distribution
     */
    public int randomGauss(int min, int max, int mean, int deviation) {
        return randomGauss(min, max, deviation, mean, false);
    }

    private int clamp(int v, int min, int max) {
        return Math.max(min, Math.min(max, v));
    }
}
