package io.reisub.openosrs.wintertodt;

import com.google.inject.Inject;
import com.google.inject.Provides;
import io.reisub.openosrs.util.CScript;
import io.reisub.openosrs.util.Util;
import io.reisub.openosrs.util.enums.Activity;
import io.reisub.openosrs.util.tasks.Eat;
import io.reisub.openosrs.wintertodt.tasks.*;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import net.runelite.api.*;
import net.runelite.api.events.*;
import net.runelite.api.widgets.Widget;
import net.runelite.client.config.ConfigManager;
import net.runelite.client.eventbus.Subscribe;
import net.runelite.client.plugins.PluginDependency;
import net.runelite.client.plugins.PluginDescriptor;
import net.runelite.client.plugins.iutils.iUtils;
import net.runelite.client.plugins.iutils.scene.Position;
import org.pf4j.Extension;

import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static net.runelite.api.AnimationID.*;
import static net.runelite.api.ItemID.BRUMA_KINDLING;
import static net.runelite.api.ItemID.BRUMA_ROOT;

@Extension
@PluginDependency(Util.class)
@PluginDependency(iUtils.class)
@PluginDescriptor(
	name = "Chaos Wintertodt",
	description = "Plays the Wintertodt minigame",
	enabledByDefault = false
)
@Slf4j
public class Wintertodt extends CScript {
	@Inject
	public Config config;

	public static final int WINTERTODT_REGION = 6462;
	public static final int WINTERTODT_HEALTH_PACKED_ID = 25952277;
	public static final int WINTERTODT_GAME_TIMER_ID = 25952259;

	@Getter
	private int respawnTimer;

	@Getter
	private int bossHealth;

	@Getter
	@Setter
	private boolean tooCold;

	@Getter
	@Setter
	private Instant lastHop;

	@Getter
	private final List<WintertodtProjectile> projectiles = new ArrayList<>();

	private Hop hopTask;
	private Scouter scouter;
	private int fmLevel, wcLevel, fletchLevel;
	private int previousTimerValue;

	@SuppressWarnings("unused")
	@Provides
	Config provideConfig(ConfigManager configManager) {
		return configManager.getConfig(Config.class);
	}

	@Override
	protected void onStart() {
		super.onStart();

		fmLevel = game.baseLevel(Skill.FIREMAKING);
		wcLevel = game.baseLevel(Skill.WOODCUTTING);
		fletchLevel = game.baseLevel(Skill.FLETCHING);

		if (config.hop()) scouter = injector.getInstance(Scouter.class);

		Eat eatTask = injector.getInstance(Eat.class);
		eatTask.setInterval(config.minEatHP(), config.maxEatHP());

		tasks.add(eatTask);
		tasks.add(kittenTask);
		if (config.dodgeProjectiles()) {
			addTask(DodgeProjectile.class);
		}
		addTask(OpenInventory.class);
		addTask(EatWhileWaiting.class);
		addTask(OpenCrates.class);
		addTask(GoToBank.class);
		addTask(HandleBank.class);
		addTask(GoToWintertodt.class);
		if (config.hop()) {
			hopTask = injector.getInstance(Hop.class);
			tasks.add(hopTask);
		}
		addTask(MoveToBrazier.class);
		addTask(Fix.class);
		addTask(Light.class);
		addTask(Fletch.class);
		addTask(ChangeSide.class);
		addTask(Burn.class);
		addTask(Chop.class);
	}

	@Override
	protected void onStop() {
		super.onStop();
		scouter = null;
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onStatChanged(StatChanged event) {
		Skill skill = event.getSkill();
		int level = event.getLevel();

		if (skill == Skill.FIREMAKING && level > fmLevel) {
			setActivity(Activity.IDLE);
			fmLevel = event.getLevel();
		} else if (skill == Skill.WOODCUTTING && level > wcLevel) {
			setActivity(Activity.IDLE);
			wcLevel = event.getLevel();
		} else if (skill == Skill.FLETCHING && level > fletchLevel) {
			setActivity(Activity.IDLE);
			fletchLevel = event.getLevel();
		} else if (skill == Skill.FIREMAKING && currentActivity == Activity.LIGHTING_BRAZIER) {
			setActivity(Activity.IDLE);
		} else if (skill == Skill.CONSTRUCTION && currentActivity == Activity.FIXING_BRAZIER) {
			setActivity(Activity.IDLE);
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onGameTick(GameTick event) {
		if (!isInWintertodtRegion() || game.client().getGameState() != GameState.LOGGED_IN) return;

		if (scouter != null) {
			scouter.onGameTick();
		}

		if (hopTask != null) {
			hopTask.onGameTick();
		}

		parseBossHealth();
		//parseRespawnTime();

		checkActionTimeout();
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onChatMessage(ChatMessage chatMessage) {
		if (!isInWintertodtRegion()) return;

		ChatMessageType chatMessageType = chatMessage.getType();
		if (chatMessageType != ChatMessageType.GAMEMESSAGE && chatMessageType != ChatMessageType.SPAM) return;

		MessageNode messageNode = chatMessage.getMessageNode();
		final InterruptType interruptType;

		if (messageNode.getValue().startsWith("You carefully fletch the root"))  {
			setActivity(Activity.FLETCHING);
			return;
		}

		if (messageNode.getValue().startsWith("The cold of")) {
			interruptType = InterruptType.COLD;
		} else if (messageNode.getValue().startsWith("The freezing cold attack")) {
			interruptType = InterruptType.SNOWFALL;
		} else if (messageNode.getValue().startsWith("The brazier is broken and shrapnel")) {
			interruptType = InterruptType.BRAZIER;
		} else if (messageNode.getValue().startsWith("You have run out of bruma roots")) {
			interruptType = InterruptType.OUT_OF_ROOTS;
		} else if (messageNode.getValue().startsWith("Your inventory is too full")) {
			interruptType = InterruptType.INVENTORY_FULL;
		} else if (messageNode.getValue().startsWith("You fix the brazier")) {
			interruptType = InterruptType.FIXED_BRAZIER;
		} else if (messageNode.getValue().startsWith("You light the brazier.")) {
			interruptType = InterruptType.LIT_BRAZIER;
		} else if (messageNode.getValue().startsWith("The brazier has gone out.")) {
			interruptType = InterruptType.BRAZIER_WENT_OUT;
		} else if (messageNode.getValue().startsWith("You eat")) {
			interruptType = InterruptType.EAT;
		} else if (messageNode.getValue().startsWith("Your hands are too cold")) {
			interruptType = InterruptType.TOO_COLD;
			tooCold = true;
		} else {
			return;
		}

		boolean wasInterrupted = false;

		switch (interruptType) {
			case EAT:
			case LIT_BRAZIER:
			case INVENTORY_FULL:
			case OUT_OF_ROOTS:
			case BRAZIER_WENT_OUT:
			case FIXED_BRAZIER:
				wasInterrupted = true;
				break;
			case COLD:
			case BRAZIER:
			case SNOWFALL:
				if (currentActivity != Activity.WOODCUTTING && currentActivity != Activity.IDLE) {
					wasInterrupted = true;
				}
				break;
		}

		if (wasInterrupted) {
			setActivity(Activity.IDLE);
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onAnimationChanged(AnimationChanged event) {
		if (game.client().getGameState() != GameState.LOGGED_IN || !isInWintertodtRegion()) return;

		if (event.getActor() != game.client().getLocalPlayer()) return;

		int animId = game.localPlayer().animation();
		switch (animId) {
			case WOODCUTTING_BRONZE:
			case WOODCUTTING_IRON:
			case WOODCUTTING_STEEL:
			case WOODCUTTING_BLACK:
			case WOODCUTTING_MITHRIL:
			case WOODCUTTING_ADAMANT:
			case WOODCUTTING_RUNE:
			case WOODCUTTING_GILDED:
			case WOODCUTTING_DRAGON:
			case WOODCUTTING_DRAGON_OR:
			case WOODCUTTING_INFERNAL:
			case WOODCUTTING_3A_AXE:
			case WOODCUTTING_CRYSTAL:
			case WOODCUTTING_TRAILBLAZER:
				setActivity(Activity.WOODCUTTING);
				break;
			case FLETCHING_BOW_CUTTING:
				setActivity(Activity.FLETCHING);
				break;
			case LOOKING_INTO:
				setActivity(Activity.FEEDING_BRAZIER);
				break;
			case FIREMAKING:
				setActivity(Activity.LIGHTING_BRAZIER);
				break;
			case CONSTRUCTION:
			case CONSTRUCTION_IMCANDO:
				setActivity(Activity.FIXING_BRAZIER);
				break;
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onItemContainerChanged(ItemContainerChanged event) {
		final ItemContainer container = event.getItemContainer();

		if (!isInWintertodtRegion() || container != game.client().getItemContainer(InventoryID.INVENTORY)) return;

		int logs = (int) game.inventory().withId(BRUMA_ROOT).count();
		int kindling = (int) game.inventory().withId(BRUMA_KINDLING).count();

		if (logs == 0 && currentActivity == Activity.FLETCHING) {
			setActivity(Activity.IDLE);
		} else if (logs == 0 && kindling == 0 && currentActivity == Activity.FEEDING_BRAZIER) {
			setActivity(Activity.IDLE);
		} else if (shouldStartFeeding() && (currentActivity == Activity.WOODCUTTING || currentActivity == Activity.FLETCHING)) {
			setActivity(Activity.IDLE);
		} else if (shouldStartFletching() && currentActivity == Activity.WOODCUTTING) {
			setActivity(Activity.IDLE);
		}
	}

	@SuppressWarnings("unused")
	@Subscribe
	private void onVarbitChanged(VarbitChanged varbitChanged) {
		int timerValue = game.client().getVar(Varbits.WINTERTODT_TIMER);
		if (timerValue != previousTimerValue) {
			respawnTimer = timerValue * 30 / 50;
			previousTimerValue = timerValue;
		}
	}

	@Subscribe
	private void onProjectileMoved(ProjectileMoved event) {
		Projectile projectile = event.getProjectile();

		if (projectile.getInteracting() != null) return;

		if (projectile.getId() == 501) {
			int x = game.client().getBaseX() + event.getPosition().getSceneX();
			int y = game.client().getBaseY() + event.getPosition().getSceneY();
			int cycles = projectile.getEndCycle() - projectile.getStartMovementCycle();

			// we don't care about any of the projectiles that don't go to our play area
			if (y >= 4001) return;

			if (cycles == 200) {
				projectiles.add(new WintertodtProjectile(x, y, true, Instant.now()));
			} else if ((x == 1638 && y == 3997) || (x == 1619 && y == 3997)) {
				if (cycles == 120) {
					projectiles.add(new WintertodtProjectile(x, y, false, Instant.now()));
				}
			}
		}
	}

	public boolean isInWintertodtRegion() {
		return game.localPlayer().position().regionID() == WINTERTODT_REGION;
	}

	public boolean bossIsUp() {
		return isInWintertodtRegion()
				&& getBossHealth() > 0
				&& getRespawnTimer() <= 0;
	}

	public boolean shouldStartFletching() {
		long rootCount = game.inventory().withName("Bruma root").count();
		long kindlingCount = game.inventory().withName("Bruma kindling").count();

		return rootCount * 2 + kindlingCount >= getBossHealth();
	}

	public boolean shouldStartFeeding() {
		long burnablesCount = game.inventory().withNamePart("Bruma").count();

		return burnablesCount >= getBossHealth();
	}

	public Side getNearestSide() {
		Position playerPos = game.localPlayer().position();
		if (playerPos.distanceTo(Side.EAST.getPositionNearBrazier()) > playerPos.distanceTo(Side.WEST.getPositionNearBrazier())) {
			return Side.WEST;
		} else {
			return Side.EAST;
		}
	}

	public Side getFurthestSide() {
		Position playerPos = game.localPlayer().position();
		if (playerPos.distanceTo(Side.EAST.getPositionNearBrazier()) < playerPos.distanceTo(Side.WEST.getPositionNearBrazier())) {
			return Side.WEST;
		} else {
			return Side.EAST;
		}
	}

	private void parseBossHealth() {
		Widget healthWidget = game.client.getWidget(WINTERTODT_HEALTH_PACKED_ID);

		if (healthWidget != null) {
			Pattern regex = Pattern.compile("\\d+");
			Matcher bossHealthMatcher = regex.matcher(healthWidget.getText());

			if (bossHealthMatcher.find()) {
				bossHealth = Integer.parseInt(bossHealthMatcher.group(0));
				if (bossHealth > 0) {
					respawnTimer = -1;
				}
			} else {
				bossHealth = -1;
			}
		}
	}

	private void parseRespawnTime() {
		Widget timerWidget = game.client.getWidget(WINTERTODT_GAME_TIMER_ID);

		if (timerWidget != null) {
			Pattern regex = Pattern.compile("\\d:\\d+");
			Matcher timerMatcher = regex.matcher(timerWidget.getText());

			if (timerMatcher.find()) {
				String[] time = timerMatcher.group(0).split(":");
				String minutes = time[0];
				String seconds = time[1];

				respawnTimer = Integer.parseInt(minutes) * 60 + Integer.parseInt(seconds);
			} else {
				respawnTimer = -1;
			}
		}
	}
}